
# returns:
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'19',
            'msvc':'19',
            'msvc_long':'19.20.27027.1',
            'vcx_tool':'15.0.25914.0',
            'platform_toolset':'142',
            'windows_sdk':'10.0'              # Windows 10 (latest installed)
            # 'windows_sdk':'10.0.17763.0'    # Windows 10 version 1809
            # 'windows_sdk':'10.0.17134.0'    # Windows 10 version 1803
            # 'windows_sdk':'10.0.16299.0'    # Windows 10 Creators Update
            # 'windows_sdk':'10.0.14393.795'  # Windows 10
          }
