/* Generated by tuneup.c, 2009-10-08, gcc 4.2 */

#define MUL_KARATSUBA_THRESHOLD          22
#define MUL_TOOM3_THRESHOLD              89
#define MUL_TOOM4_THRESHOLD             202
#define MUL_TOOM7_THRESHOLD             286

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          42
#define SQR_TOOM3_THRESHOLD             187
#define SQR_TOOM4_THRESHOLD             232
#define SQR_TOOM7_THRESHOLD             680

#define MULLOW_BASECASE_THRESHOLD         6
#define MULLOW_DC_THRESHOLD              88
#define MULLOW_MUL_THRESHOLD           6713

#define MULHIGH_BASECASE_THRESHOLD        8
#define MULHIGH_DC_THRESHOLD             87
#define MULHIGH_MUL_THRESHOLD          6713

#define MULMOD_2EXPM1_THRESHOLD          18

#define POWM_THRESHOLD                  185
#define FAC_UI_THRESHOLD               2169

#define GCDEXT_THRESHOLD                 82
#define JACOBI_BASE_METHOD                1

#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1
#define DIVREM_2_THRESHOLD                0  /* always */
#define DIVEXACT_1_THRESHOLD              0  /* always (native) */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always */
#define MOD_1_1_THRESHOLD                 7
#define MOD_1_2_THRESHOLD                 8
#define MOD_1_3_THRESHOLD               996
#define DIVREM_HENSEL_QR_1_THRESHOLD      9
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD      3
#define DIVREM_EUCLID_HENSEL_THRESHOLD     59

#define ROOTREM_THRESHOLD                 6

#define GET_STR_DC_THRESHOLD             25
#define GET_STR_PRECOMPUTE_THRESHOLD     36

#define MUL_FFT_FULL_THRESHOLD              7168

#define SQR_FFT_FULL_THRESHOLD              3840

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 2 }, { 2, 1 }, { 2, 1 }, { 1, 0 } }

#define MULMOD_TAB \
   { 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 15

#define FFT_MULMOD_2EXPP1_CUTOFF 128

/* Tuneup completed successfully, took 62 seconds */
