
# returns:
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'2012',
            'msvc':'11',
            'msvc_long':'11',
            'vcx_tool':'4.0',
            'platform_toolset':'110',
            'windows_sdk':'8.1',
            # 'windows_sdk':'7.1.51106'
          }
