// Copyright 2017 Elias Kosunen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// This file is a part of scnlib:
//     https://github.com/eliaskosunen/scnlib

#include <scn/scn.h>
#include <iostream>

int main()
{
    auto source = std::string{"First line\nSecond line"};
    std::string line{};
    if (auto ret = scn::getline(source, line)) {
        std::cout << "First line was: '" << line << "'\n";
        if ((ret = scn::getline(ret.range(), line))) {
            std::cout << "Second line was: '" << line << "'\n";
        }
    }
}
