/* Generated by tuneup.c, 2014-03-24, gcc 4.8 */

#define MUL_KARATSUBA_THRESHOLD          24
#define MUL_TOOM3_THRESHOLD              93
#define MUL_TOOM4_THRESHOLD             262
#define MUL_TOOM8H_THRESHOLD            262

#define SQR_BASECASE_THRESHOLD           16
#define SQR_KARATSUBA_THRESHOLD          66
#define SQR_TOOM3_THRESHOLD             107
#define SQR_TOOM4_THRESHOLD             292
#define SQR_TOOM8_THRESHOLD             292

#define POWM_THRESHOLD                  110

#define DIVREM_1_NORM_THRESHOLD           0  /* always */
#define DIVREM_1_UNNORM_THRESHOLD         0  /* always */
#define MOD_1_NORM_THRESHOLD              0  /* always */
#define MOD_1_UNNORM_THRESHOLD            0  /* always */
#define USE_PREINV_DIVREM_1               1
#define USE_PREINV_MOD_1                  1
#define DIVREM_2_THRESHOLD                0  /* always */
#define DIVEXACT_1_THRESHOLD              0  /* always */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */
#define MOD_1_1_THRESHOLD                 6
#define MOD_1_2_THRESHOLD                 8
#define MOD_1_3_THRESHOLD                29
#define DIVREM_HENSEL_QR_1_THRESHOLD     31
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD      9
#define DIVREM_EUCLID_HENSEL_THRESHOLD     75

#define MUL_FFT_FULL_THRESHOLD         4064

#define SQR_FFT_FULL_THRESHOLD         4000

#define MULLOW_BASECASE_THRESHOLD         0  /* always */
#define MULLOW_DC_THRESHOLD              48
#define MULLOW_MUL_THRESHOLD           2562

#define MULHIGH_BASECASE_THRESHOLD        4
#define MULHIGH_DC_THRESHOLD             20
#define MULHIGH_MUL_THRESHOLD          2562

#define MULMOD_2EXPM1_THRESHOLD          24

#define DC_DIV_QR_THRESHOLD              56
#define INV_DIV_QR_THRESHOLD           1894
#define INV_DIVAPPR_Q_N_THRESHOLD        10
#define DC_DIV_Q_THRESHOLD               68
#define DC_DIVAPPR_Q_THRESHOLD           95
#define INV_DIV_QR_THRESHOLD           1894
#define DC_BDIV_QR_THRESHOLD             77
#define DC_BDIV_Q_THRESHOLD              34

#define ROOTREM_THRESHOLD                11

#define MATRIX22_STRASSEN_THRESHOLD      19
#define HGCD_THRESHOLD                  105
#define HGCD_APPR_THRESHOLD             133
#define HGCD_REDUCE_THRESHOLD          6852
#define GCD_DC_THRESHOLD                555
#define GCDEXT_DC_THRESHOLD             427
#define JACOBI_BASE_METHOD                3

#define GET_STR_DC_THRESHOLD             10
#define GET_STR_PRECOMPUTE_THRESHOLD     22
#define SET_STR_DC_THRESHOLD            704
#define SET_STR_PRECOMPUTE_THRESHOLD   1820

#define FAC_DSC_THRESHOLD               141
#define FAC_ODD_THRESHOLD                29

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 2 }, { 2, 1 }, { 2, 0 }, { 1, 0 } }

#define MULMOD_TAB \
   { 4, 3, 3, 3, 3, 2, 2, 3, 2, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 15

#define FFT_MULMOD_2EXPP1_CUTOFF 256


/* Tuneup completed successfully, took 157 seconds */

