/* Generated by tuneup.c, 2017-02-14, gcc 5.4 */

#define MUL_KARATSUBA_THRESHOLD          19
#define MUL_TOOM3_THRESHOLD             137
#define MUL_TOOM4_THRESHOLD             193
#define MUL_TOOM8H_THRESHOLD            381

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          32
#define SQR_TOOM3_THRESHOLD              89
#define SQR_TOOM4_THRESHOLD             258
#define SQR_TOOM8_THRESHOLD             494

#define DIVREM_1_NORM_THRESHOLD       MP_SIZE_T_MAX  /* never */
#define DIVREM_1_UNNORM_THRESHOLD     MP_SIZE_T_MAX  /* never */
#define MOD_1_NORM_THRESHOLD              0  /* always */
#define MOD_1_UNNORM_THRESHOLD            0  /* always */
#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1
#define DIVEXACT_1_THRESHOLD              0  /* always */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */
#define MOD_1_1_THRESHOLD                 6
#define MOD_1_2_THRESHOLD                10
#define MOD_1_3_THRESHOLD                17
#define DIVREM_HENSEL_QR_1_THRESHOLD     31
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD      6
#define DIVREM_EUCLID_HENSEL_THRESHOLD     35

#define MUL_FFT_FULL_THRESHOLD         5056

#define SQR_FFT_FULL_THRESHOLD         3008

#define MULLOW_BASECASE_THRESHOLD         0  /* always */
#define MULLOW_DC_THRESHOLD              21
#define MULLOW_MUL_THRESHOLD           3369

#define MULMID_TOOM42_THRESHOLD          24

#define MULHIGH_BASECASE_THRESHOLD       10
#define MULHIGH_DC_THRESHOLD             22
#define MULHIGH_MUL_THRESHOLD          3336

#define MULMOD_2EXPM1_THRESHOLD          18

#define DC_DIV_QR_THRESHOLD              29
#define INV_DIV_QR_THRESHOLD           2444
#define INV_DIVAPPR_Q_N_THRESHOLD        29
#define DC_DIV_Q_THRESHOLD               45
#define INV_DIV_Q_THRESHOLD            1258
#define DC_DIVAPPR_Q_THRESHOLD          104
#define INV_DIVAPPR_Q_THRESHOLD       14091
#define DC_BDIV_QR_THRESHOLD             62
#define DC_BDIV_Q_THRESHOLD              96

#define BINV_NEWTON_THRESHOLD             8
#define REDC_1_TO_REDC_2_THRESHOLD       22
#define REDC_2_TO_REDC_N_THRESHOLD       36

#define ROOTREM_THRESHOLD                 6

#define MATRIX22_STRASSEN_THRESHOLD      15
#define HGCD_THRESHOLD                  103
#define HGCD_APPR_THRESHOLD              98
#define HGCD_REDUCE_THRESHOLD          6852
#define GCD_DC_THRESHOLD                752
#define GCDEXT_DC_THRESHOLD             496
#define JACOBI_BASE_METHOD                2

#define GET_STR_DC_THRESHOLD             14
#define GET_STR_PRECOMPUTE_THRESHOLD     20
#define SET_STR_DC_THRESHOLD           2324
#define SET_STR_PRECOMPUTE_THRESHOLD   3246

#define FAC_DSC_THRESHOLD               858
#define FAC_ODD_THRESHOLD                44

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 4 }, { 4, 3 }, { 3, 2 }, { 3, 2 }, { 2, 1 } }

#define MULMOD_TAB \
   { 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 19

#define FFT_MULMOD_2EXPP1_CUTOFF 256


/* Tuneup completed successfully, took 63 seconds */