/* Generated by tuneup.c, 2014-03-24, gcc 4.1 */

#define MUL_KARATSUBA_THRESHOLD          36
#define MUL_TOOM3_THRESHOLD             117
#define MUL_TOOM4_THRESHOLD             312
#define MUL_TOOM8H_THRESHOLD            312

#define SQR_BASECASE_THRESHOLD           11
#define SQR_KARATSUBA_THRESHOLD          80
#define SQR_TOOM3_THRESHOLD             131
#define SQR_TOOM4_THRESHOLD             582
#define SQR_TOOM8_THRESHOLD             582

#define POWM_THRESHOLD                  984

#define DIVREM_1_NORM_THRESHOLD           9
#define DIVREM_1_UNNORM_THRESHOLD     MP_SIZE_T_MAX  /* never */
#define MOD_1_NORM_THRESHOLD              9
#define MOD_1_UNNORM_THRESHOLD        MP_SIZE_T_MAX  /* never */
#define USE_PREINV_DIVREM_1               1
#define USE_PREINV_MOD_1                  1
#define DIVREM_2_THRESHOLD                9
#define DIVEXACT_1_THRESHOLD              0  /* always */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always */
#define MOD_1_1_THRESHOLD                 8
#define MOD_1_2_THRESHOLD                 8
#define MOD_1_3_THRESHOLD                25
#define DIVREM_HENSEL_QR_1_THRESHOLD    996
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD    996
#define DIVREM_EUCLID_HENSEL_THRESHOLD    100

#define MUL_FFT_FULL_THRESHOLD         2752

#define SQR_FFT_FULL_THRESHOLD         1952

#define MULLOW_BASECASE_THRESHOLD         0  /* always */
#define MULLOW_DC_THRESHOLD             122
#define MULLOW_MUL_THRESHOLD           3336

#define MULHIGH_BASECASE_THRESHOLD        0  /* always */
#define MULHIGH_DC_THRESHOLD            118
#define MULHIGH_MUL_THRESHOLD          8942

#define MULMOD_2EXPM1_THRESHOLD          20

#define DC_DIV_QR_THRESHOLD             146
#define INV_DIV_QR_THRESHOLD           3547
#define INV_DIVAPPR_Q_N_THRESHOLD       146
#define DC_DIV_Q_THRESHOLD              225
#define INV_DIV_Q_THRESHOLD            2172
#define DC_DIVAPPR_Q_THRESHOLD          351
#define INV_DIVAPPR_Q_THRESHOLD       18999
#define DC_BDIV_QR_THRESHOLD            138
#define DC_BDIV_Q_THRESHOLD             209

#define ROOTREM_THRESHOLD                 6

#define MATRIX22_STRASSEN_THRESHOLD      18
#define HGCD_THRESHOLD                  193
#define HGCD_APPR_THRESHOLD             296
#define HGCD_REDUCE_THRESHOLD          6852
#define GCD_DC_THRESHOLD                634
#define GCDEXT_DC_THRESHOLD             622
#define JACOBI_BASE_METHOD                3

#define GET_STR_DC_THRESHOLD             11
#define GET_STR_PRECOMPUTE_THRESHOLD     18
#define SET_STR_DC_THRESHOLD            866
#define SET_STR_PRECOMPUTE_THRESHOLD   2296

#define FAC_DSC_THRESHOLD               232
#define FAC_ODD_THRESHOLD                24

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 3, 3 }, { 3, 2 }, { 2, 1 }, { 1, 0 }, { 0, 0 } }

#define MULMOD_TAB \
   { 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 19

#define FFT_MULMOD_2EXPP1_CUTOFF 128


/* Tuneup completed successfully, took 3405 seconds */

