#pragma once

// I DO NOT TAKE CREDIT FOR THIS CODE -- I SIMPLY MODIFIED IT

// These are needed for some functionality
#include "../internal/config.hpp"
#include "../utils/time.hpp"
#include "../math/coreMath.hpp"

#include "inplace/reducedMath.hpp"
#include "inplace/index.hpp"
#include "inplace/transpose.hpp"
#include "float16/float16.hpp"
