/* Generated by tuneup.c, 2010-03-24, gcc 4.3 */

#define MUL_KARATSUBA_THRESHOLD          26
#define MUL_TOOM3_THRESHOLD              97
#define MUL_TOOM4_THRESHOLD             230
#define MUL_TOOM8H_THRESHOLD            254

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          50
#define SQR_TOOM3_THRESHOLD              95
#define SQR_TOOM4_THRESHOLD             256
#define SQR_TOOM8_THRESHOLD             262

#define POWM_THRESHOLD                  180

#define GCD_THRESHOLD                   330
#define GCDEXT_THRESHOLD                996
#define JACOBI_BASE_METHOD                1

#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1  /* native */
#define DIVREM_2_THRESHOLD                0  /* always */
#define DIVEXACT_1_THRESHOLD              0  /* always (native) */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */
#define MOD_1_1_THRESHOLD                10
#define MOD_1_2_THRESHOLD                71
#define MOD_1_3_THRESHOLD                72
#define DIVREM_HENSEL_QR_1_THRESHOLD      5
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD      5
#define DIVREM_EUCLID_HENSEL_THRESHOLD      8

#define ROOTREM_THRESHOLD                 6

#define GET_STR_DC_THRESHOLD             15
#define GET_STR_PRECOMPUTE_THRESHOLD     26
#define SET_STR_DC_THRESHOLD          10475
#define SET_STR_PRECOMPUTE_THRESHOLD  10475

#define MUL_FFT_FULL_THRESHOLD         1568

#define SQR_FFT_FULL_THRESHOLD         1568

#define MULLOW_BASECASE_THRESHOLD         8
#define MULLOW_DC_THRESHOLD              45
#define MULLOW_MUL_THRESHOLD           3436

#define MULHIGH_BASECASE_THRESHOLD       17
#define MULHIGH_DC_THRESHOLD             21
#define MULHIGH_MUL_THRESHOLD          3436

#define MULMOD_2EXPM1_THRESHOLD          20

#define FAC_UI_THRESHOLD               6906
#define DC_DIV_QR_THRESHOLD              92
#define DC_DIVAPPR_Q_N_THRESHOLD        807
#define INV_DIV_QR_THRESHOLD           2801
#define INV_DIVAPPR_Q_N_THRESHOLD       807
#define DC_DIV_Q_THRESHOLD              807
#define INV_DIV_Q_THRESHOLD            1164
#define DC_DIVAPPR_Q_THRESHOLD          807
#define INV_DIVAPPR_Q_THRESHOLD        3690
#define DC_BDIV_QR_THRESHOLD             88
#define DC_BDIV_Q_THRESHOLD             448

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 2 }, { 2, 1 }, { 2, 1 }, { 1, 0 } }

#define MULMOD_TAB \
   { 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 15

#define FFT_MULMOD_2EXPP1_CUTOFF 128

/* Tuneup completed successfully, took 1025 seconds */
