/* Generated by tuneup.c, 2010-03-24, gcc 4.4 */

#define MUL_KARATSUBA_THRESHOLD          22
#define MUL_TOOM3_THRESHOLD             129
#define MUL_TOOM4_THRESHOLD             181
#define MUL_TOOM8H_THRESHOLD            226

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          40
#define SQR_TOOM3_THRESHOLD             126
#define SQR_TOOM4_THRESHOLD             214
#define SQR_TOOM8_THRESHOLD             222

#define POWM_THRESHOLD                  984

#define GCDEXT_THRESHOLD                104
#define JACOBI_BASE_METHOD                1

#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1
#define DIVREM_2_THRESHOLD                0  /* always */
#define DIVEXACT_1_THRESHOLD              0  /* always (native) */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always */
#define MOD_1_1_THRESHOLD                 6
#define MOD_1_2_THRESHOLD                 6
#define MOD_1_3_THRESHOLD                19
#define DIVREM_HENSEL_QR_1_THRESHOLD    996
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD    996
#define DIVREM_EUCLID_HENSEL_THRESHOLD    399

#define ROOTREM_THRESHOLD                 6

#define GET_STR_DC_THRESHOLD             16
#define GET_STR_PRECOMPUTE_THRESHOLD     26
#define SET_STR_DC_THRESHOLD          14963
#define SET_STR_PRECOMPUTE_THRESHOLD  16195

#define MUL_FFT_FULL_THRESHOLD         1664

#define SQR_FFT_FULL_THRESHOLD         1664

#define MULLOW_BASECASE_THRESHOLD         7
#define MULLOW_DC_THRESHOLD              93
#define MULLOW_MUL_THRESHOLD           2852

#define MULHIGH_BASECASE_THRESHOLD        8
#define MULHIGH_DC_THRESHOLD             87
#define MULHIGH_MUL_THRESHOLD          2852

#define MULMOD_2EXPM1_THRESHOLD          18

#define FAC_UI_THRESHOLD               3300
#define DC_DIV_QR_THRESHOLD              60
#define DC_DIVAPPR_Q_N_THRESHOLD        195
#define INV_DIV_QR_THRESHOLD           2857
#define INV_DIVAPPR_Q_N_THRESHOLD       195
#define DC_DIV_Q_THRESHOLD              229
#define INV_DIV_Q_THRESHOLD            2350
#define DC_DIVAPPR_Q_THRESHOLD          195
#define INV_DIVAPPR_Q_THRESHOLD        5363
#define DC_BDIV_QR_THRESHOLD            110
#define DC_BDIV_Q_THRESHOLD              42

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 2 }, { 2, 1 }, { 2, 1 }, { 1, 0 } }

#define MULMOD_TAB \
   { 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 15

#define FFT_MULMOD_2EXPP1_CUTOFF 128

/* Tuneup completed successfully, took 404 seconds */
