/* Generated by tuneup.c, 2014-03-21, gcc 4.4 */

#define MUL_KARATSUBA_THRESHOLD          14
#define MUL_TOOM3_THRESHOLD              81
#define MUL_TOOM4_THRESHOLD             196
#define MUL_TOOM8H_THRESHOLD            220

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          32
#define SQR_TOOM3_THRESHOLD             178
#define SQR_TOOM4_THRESHOLD             296
#define SQR_TOOM8_THRESHOLD             300

#define POWM_THRESHOLD                  287

#define DIVREM_1_NORM_THRESHOLD       MP_SIZE_T_MAX  /* never */
#define DIVREM_1_UNNORM_THRESHOLD     MP_SIZE_T_MAX  /* never */
#define MOD_1_NORM_THRESHOLD              0  /* always */
#define MOD_1_UNNORM_THRESHOLD            0  /* always */
#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1
#define DIVEXACT_1_THRESHOLD              0  /* always */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */
#define MOD_1_1_THRESHOLD                 6
#define MOD_1_2_THRESHOLD                 9
#define MOD_1_3_THRESHOLD                20
#define DIVREM_HENSEL_QR_1_THRESHOLD      9
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD      6
#define DIVREM_EUCLID_HENSEL_THRESHOLD     39

#define MUL_FFT_FULL_THRESHOLD         3008

#define SQR_FFT_FULL_THRESHOLD         2496

#define MULLOW_BASECASE_THRESHOLD         4
#define MULLOW_DC_THRESHOLD              13
#define MULLOW_MUL_THRESHOLD           4525

#define MULHIGH_BASECASE_THRESHOLD        8
#define MULHIGH_DC_THRESHOLD             23
#define MULHIGH_MUL_THRESHOLD            28

#define MULMOD_2EXPM1_THRESHOLD          14

#define DC_DIV_QR_THRESHOLD              59
#define INV_DIV_QR_THRESHOLD            456
#define INV_DIVAPPR_Q_N_THRESHOLD        59
#define DC_DIV_Q_THRESHOLD               63
#define INV_DIV_Q_THRESHOLD            1589
#define DC_DIVAPPR_Q_THRESHOLD           45
#define INV_DIVAPPR_Q_THRESHOLD       17642
#define DC_BDIV_QR_THRESHOLD             36
#define DC_BDIV_Q_THRESHOLD              42

#define ROOTREM_THRESHOLD                 6

#define MATRIX22_STRASSEN_THRESHOLD      26
#define HGCD_THRESHOLD                   92
#define HGCD_APPR_THRESHOLD             109
#define HGCD_REDUCE_THRESHOLD            45
#define GCD_DC_THRESHOLD                102
#define GCDEXT_DC_THRESHOLD             108
#define JACOBI_BASE_METHOD                2

#define GET_STR_DC_THRESHOLD             13
#define GET_STR_PRECOMPUTE_THRESHOLD     36
#define SET_STR_DC_THRESHOLD            486
#define SET_STR_PRECOMPUTE_THRESHOLD   1659

#define FAC_DSC_THRESHOLD               486
#define FAC_ODD_THRESHOLD                27

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 3 }, { 3, 1 }, { 1, 1 }, { 1, 1 } }

#define MULMOD_TAB \
   { 4, 4, 3, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 19

#define FFT_MULMOD_2EXPP1_CUTOFF 128


/* Tuneup completed successfully, took 505 seconds */

