/* Generated by tuneup.c, 2015-09-30, gcc 4.8 */

#define MUL_KARATSUBA_THRESHOLD          30
#define MUL_TOOM3_THRESHOLD              81
#define MUL_TOOM4_THRESHOLD             238
#define MUL_TOOM8H_THRESHOLD            354

#define SQR_BASECASE_THRESHOLD            0  /* always (native) */
#define SQR_KARATSUBA_THRESHOLD          36
#define SQR_TOOM3_THRESHOLD             138
#define SQR_TOOM4_THRESHOLD             387
#define SQR_TOOM8_THRESHOLD             387

#define POWM_THRESHOLD                  984

#define USE_PREINV_DIVREM_1               1  /* native */
#define USE_PREINV_MOD_1                  1  /* native */
#define DIVREM_2_THRESHOLD               11
#define DIVEXACT_1_THRESHOLD              0  /* always (native) */
#define MODEXACT_1_ODD_THRESHOLD          0  /* always (native) */
#define MOD_1_1_THRESHOLD                83
#define MOD_1_2_THRESHOLD               191
#define MOD_1_3_THRESHOLD               850
#define DIVREM_HENSEL_QR_1_THRESHOLD    996
#define RSH_DIVREM_HENSEL_QR_1_THRESHOLD    996
#define DIVREM_EUCLID_HENSEL_THRESHOLD     24

#define MUL_FFT_FULL_THRESHOLD         5696

#define SQR_FFT_FULL_THRESHOLD         4064

#define MULLOW_BASECASE_THRESHOLD        11
#define MULLOW_DC_THRESHOLD              99
#define MULLOW_MUL_THRESHOLD           3176

#define MULHIGH_BASECASE_THRESHOLD       11
#define MULHIGH_DC_THRESHOLD             95
#define MULHIGH_MUL_THRESHOLD          2966

#define MULMOD_2EXPM1_THRESHOLD          24

#define DC_DIV_QR_THRESHOLD              31
#define INV_DIV_QR_THRESHOLD           1970
#define INV_DIVAPPR_Q_N_THRESHOLD        31
#define DC_DIV_Q_THRESHOLD               45
#define INV_DIV_Q_THRESHOLD            1078
#define DC_DIVAPPR_Q_THRESHOLD           27
#define INV_DIVAPPR_Q_THRESHOLD        5025
#define DC_BDIV_QR_THRESHOLD            102
#define DC_BDIV_Q_THRESHOLD              21

#define ROOTREM_THRESHOLD                 6

#define MATRIX22_STRASSEN_THRESHOLD      23
#define HGCD_THRESHOLD                   87
#define HGCD_APPR_THRESHOLD              89
#define HGCD_REDUCE_THRESHOLD          6852
#define GCD_DC_THRESHOLD                618
#define GCDEXT_DC_THRESHOLD             584
#define JACOBI_BASE_METHOD                2

#define GET_STR_DC_THRESHOLD             10
#define GET_STR_PRECOMPUTE_THRESHOLD     20
#define SET_STR_DC_THRESHOLD            272
#define SET_STR_PRECOMPUTE_THRESHOLD   1330

#define FAC_DSC_THRESHOLD               164
#define FAC_ODD_THRESHOLD                30

/* fft_tuning -- autogenerated by tune-fft */

#define FFT_TAB \
   { { 4, 3 }, { 3, 2 }, { 2, 2 }, { 2, 1 }, { 1, 0 } }

#define MULMOD_TAB \
   { 4, 3, 3, 3, 3, 2, 2, 3, 3, 2, 2, 2, 2, 1, 1 }

#define FFT_N_NUM 15

#define FFT_MULMOD_2EXPP1_CUTOFF 256


/* Tuneup completed successfully, took 151 seconds */

